/*
 * SpringProx or K531 SDK :: direct/binary.c
 * ------------------------------------------
 *
 * (Minimal) implementation of K531 or SpringProx binary protocl
 *
 * IMPORTANT NOTICE:
 *     This implementation works only for payload size < 128 bytes
 */

/*

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
  
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ref_direct.h"

unsigned char Sequence = 0;

/*
 * BinarySend
 * ----------
 * Format and send a frame to reader according to binary protocol
 * Returns size of payload on success, error code otherwise
 *
 * Payload is [CMD CODE] [ .... OPTIONAL DATA .... ]
 */
int BinarySend(unsigned char *payload, int payload_sz)
{
  int e, i;
  unsigned char b, c = 0; /* c will be the CRC */

  if (TraceReader)
  {
    printf("< %02X ", payload[0]);
    printf("%02X ", payload_sz);
    for (i=1; i<payload_sz; i++)
      printf("%02X", payload[i]);
    printf("\n");
  }

  if (payload_sz >= 128) return E_INTERNAL;

  /* Send SYN */
  b = 0x16;
  e = SerialSend(&b, 1);
  if (e != 1) return e; /* Communication error */

  /* Send dummy sequence number */
  b = Sequence++;
  e = SerialSend(&b, 1);
  if (e != 1) return e; /* Communication error */
  c ^= b;

  /* Send command code (first byte of payload) */
  e = SerialSend(&payload[0], 1);
  if (e != 1) return e; /* Communication error */
  c ^= payload[0];

  /* Send length of data (this is size of payload minus 1 for the command code) */
  b = (unsigned char) (payload_sz - 1);
  e = SerialSend(&b, 1);
  if (e != 1) return e; /* Communication error */
  c ^= b;

  /* Send data at once */
  if (payload_sz > 1)
  {
    e = SerialSend(&payload[1], payload_sz-1);
    if (e != (payload_sz-1)) return e; /* Communication error */

    for (i = 1; i < payload_sz; i++)
      c ^= payload[i];
  }

  /* Send CRC */
  e = SerialSend(&c, 1);
  if (e != 1) return e; /* Communication error */

  /* OK */
  return payload_sz;
}

/*
 * BinaryRecv
 * ----------
 * Receive and un-format a frame from reader according to binary protocol
 * Returns actual size of payload on success, error code otherwise
 *
 * Payload is [RESP CODE] [ .... OPTIONAL DATA .... ]
 */
int BinaryRecv(unsigned char *payload, int max_payload_sz)
{
  int e, i, payload_sz;
  unsigned char b, c = 0; /* c will be the CRC */

  /* Wait for SYN character */
  do
  {
    e = SerialRecv(&b, 1);
    if (e < 0)  return e;
    if (e == 0) return E_READER_MUTE;
  } while (b != 0x16);

  /* Receive sequence number */
  e = SerialRecv(&b, 1);
  if (e < 0)  return e;
  if (e == 0) return E_READER_STOP;
  c ^= b;

  /* TODO : check that sequence number matches the one sent */

  /* Receive response code code */
  e = SerialRecv(&b, 1);
  if (e < 0)  return e;
  if (e == 0) return E_READER_STOP;
  c ^= b;

  /* Small hack to detect obsolete products */
  if (b == 100)
  {
    printf("\n");
    printf("\t!!!    Panic: reader returned code -100 ('not yet implemented')\n");
    printf("\tUpgrade your reader with a newest firmware for correct operation !!!\n");
    printf("\n");
  }

  /* Put the response code in buffer (if room for it) */
  if ((payload != NULL) && (1 <= max_payload_sz))
    payload[0] = b;

  /* Receive length of payload data */
  e = SerialRecv(&b, 1);
  if (e < 0)  return e;
  if (e == 0) return E_READER_STOP;
  c ^= b;

  payload_sz = b;

  /* Receive payload data in a loop */
  for (i=1; i<=payload_sz; i++)
  {
    /* Get next byte */
    e = SerialRecv(&b, 1);
    if (e < 0)  return e;
    if (e == 0) return E_READER_STOP;
    c ^= b;

    /* Put it in buffer (if room for it) */
    if ((payload != NULL) && (i < max_payload_sz))
      payload[i] = b;
  }

  /* Recv CRC */
  e = SerialRecv(&b, 1);
  if (e < 0)  return e;
  if (e == 0) return E_READER_STOP;

  /* Check CRC */
  if (b != c) return E_FRAME_BAD_CRC;

  /* See if the response has overflowed the buffer */
  if ((payload != NULL) && (i > max_payload_sz)) return E_OVERFLOW; 

  if (TraceReader && (payload != NULL))
  {
    printf("> %02X ", payload[0]);
    printf("%02X ", payload_sz);
    for (i=1; i<=payload_sz; i++)
      printf("%02X", payload[i]);
    printf("\n");
  }

  /* OK, return actual length */
  return i;
}
