#include "ref_direct.h"
#ifdef STUB_TARGET

/*
 * This is a skeleton of the serial part for MCU
 *
 * We choose
 * 1. To send frames (MCU->K531) synchronously (SerialSend is a busy loop)
 * 2. To receive bytes (K531->MCU) through an interrupt handler.
 *    In this example SerialRecv waits until one complete frame has been
 *    received, on a 'multi-task' system it is easy to call SerialRecv
 *    only once the buffer has been fulled by the interrupt handler.
 * 
 * To detect communication timeout we need a timer (10ms resolution is OK).
 *
 * In this example we size the receiving buffer (UartRecvPipe) to its maximum
 * usable value (276 bytes). Depending on the function used by the upper-level
 * application, it is possible to reduce this size to 68 bytes only (only T=CL
 * communication with 14443-4 cards, or Mifare operation with 4k cards, need
 * more than 68 bytes).
 */

/* The receiving buffer */
/* -------------------- */
unsigned char UartRecvPipe[276];
// NOTE : sizeof(UartRecvPipe) > 255, so we need sizeof(int) >= 2
volatile unsigned int UartRecvProducer = 0;
volatile unsigned int UartRecvConsumer = 0;

/* Prototypes for timer related function, to be written elsewhere... */
void TimeoutInit(unsigned int Duration); // Duration in milli-seconds
void TimeoutCancel(void); // Stop the timer
int TimeoutExpired(); // Returns TRUE if the timeout has expired



int SerialOpen(char *DeviceName)
{
  /* Always OK (don't even care for name) */
  return 1; 
}

void SerialClose(void)
{
  /* Just do nothing */
}

/*
 * UART receive interrupt (RX full IRQ)
 * ------------------------------------
 */
[-- RX FULL INTERRUPT HANDLER --]
{
  UartRecvPipe[UartRecvProducer++] = [-- UART_RX_BYTE --];
  UartRecvProducer %= sizeof(UartRecvPipe);
}

/*
 * Exported receive function
 * -------------------------
 * Wait until the expected count of bytes has arrived,
 * or a timeout has occured (timeout is 25ms)
 * Returns the number of bytes received
 */
int SerialRecv(unsigned char *Buffer, int Size)
{
  int Count = 0;  
  
  for (Count = 0; Count < Size; Count++)
  {
    /* 25ms between 2 consecutive bytes */
    TimeoutInit(25);
        
    while (UartRecvConsumer == UartRecvProducer)
    {
      /* Nothing to do but wait until next character arrives */
      if (TimeoutExpired())
      {
        TimeoutCancel();
        return Count; /* Count<Size ==> Error */
      }
    }
    
    Buffer[Count] = UartRecvPipe[UartRecvConsumer++];
    UartRecvConsumer %= sizeof(UartRecvPipe);    
  }
  
  TimeoutCancel();
  return Size; /* OK */
}

/*
 * Send function
 * -------------
 * This one is a busy loop, can be implemented on TX empty IRQ
 */
int SerialSend(unsigned char *Buffer, int Size)
{    
  int Count;
  
  for (Count = 0; Count < Size; Count++)
  {
    /* Wait until last byte has been transmitted */
    while ( [-- UART_TX_BUSY --] )
    {
      [-- WAIT A LITTLE OR DO SOMETHING ELSE --]
    }
    /* Transmit next byte */
    [-- UART_RX_BYTE --] = Buffer[Count];
  }

  /* Always OK ! (TODO : handle UART error IRQ or error flag) */
  return Size;
}



#endif
