/*
 * SpringProx or K531 SDK :: direct/mif_ul.c
 * -----------------------------------------
 *
 * Implementation of Mifare UltraLight read & write functions
 *
 */

/*

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
  
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ref_direct.h"

/*
 * MifareUlProc
 * ------------
 *
 * The function waits for a Mifare UltraLight card, then read it
 * 4 pages at once, and display the data page by page.
 *
 * Writing into the card is done on a single page basis.
 *
 */
int MifareUlProc(void)
{
  int e, i;
  unsigned char p4, p;

  unsigned char atq[2];  /* Answer To Query for 144443 A     */
  unsigned char uid[12]; /* Serial number (UID) for 144443 A */
  unsigned char sak[1];  /* Select AcKnowledge for 144443 A  */

  unsigned char data[16];
  
  /* Infinite loop, use CTRL+C to exit */
  printf("Waiting for a Mifare UltraLight card\n");
  printf("Put your card on the reader, or hit CTRL+C to exit\n");

  for (;;)
  {
    e = SelectTypeA(atq, uid, sak);
    if (e < 0) return e; /* Fatal error */
    if (e == 0)
    {
      /* Found a type A card */
      /* ------------------- */

      /* What kind of card ? (See NXP litterature for reference) */
      if ((atq[1] == 0x00) && (atq[0] == 0x44))
      {
        /* Mifare UltraLight (or compliant ?) */
        printf("Found a Mifare UltraLight\n");
        break;
      } else
      {
        /* Mifare Classic 1k (or compliant ?) */
        printf("Found a card, but this is not a Mifare UltraLight (ATQ=%02X%02X)\n", atq[1], atq[0]);
      }

      /* We HALT the card to allow anticollision (next loop will return next card, or none) */
      e = Halt();
      if (e < 0) return e;
    }
  }

  printf("Reading 16 pages...\n");
  printf("PAGE  DATA\n");

  for (p4=0; p4<16; p4+=4)
  {
    e = MifareUlRead4Pages(p4, data);
    if (e < 0) return e; /* Fatal error */

    if (e == 0)
    {
      /* Success, let's show the data */
      for (p=0; p<4; p++)
      {
        /* Show page number */
        printf("[%02d]  ", p4+p);

        /* Show data in HEX */
        for (i=0; i<4; i++)
          printf("%02X ", data[4*p + i]);

        printf("  ");

        /* Show plain data */
        for (i=0; i<4; i++)
          printf("%c", data[4*p + i] >= ' ' ? data[4*p + i] : '.');

        printf("\n");
      }      
    } else
    {
      /* Failure, show the error code */
      printf("[%02d]  Error %d\n\t(%s)\n\n", p4, e, ErrorMessage(e));

      if (e == S_NO_CARD) return e; /* Exit if card removed, this is a permanent error */
    }
  }

  /* Done ! */
  return 0;
}

/*
 * MifareUlReadPages
 * -----------------
 * Read 4 pages (16 bytes) from a Mifare UltraLight
 */
int MifareUlRead4Pages(unsigned char addr, unsigned char data[16])
{
  unsigned char cmd[4];
  unsigned char rsp[1 + 16];    /* 1 for status, 16 for data */
  int r;

  /* Build the command */
  cmd[0] = 0x8E;    /* SPROX_PICCCOMMONREAD */
  cmd[1] = 0x30;    /* Mifare UL read command */
  cmd[2] = addr;    /* Address of first page */
  cmd[3] = 16;      /* Size of expected response */
  
  /* Send the command */
  r = Send(cmd, 4);
  if (r < 0) return r;

  /* Get the answer */
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] == 0)
  {
    if (r < 16+1) return E_FRAME_BAD_LEN;

    /* Retrieve data */
    if (data != NULL)
      memcpy(data, &rsp[1], 16);

    /* OK */
    return 0;
  }

  /* Error ? */
  if (rsp[0] >= 26)
    return E_READER_FATAL;
  if (rsp[0] == 1)
    return S_NO_CARD;
  if (rsp[0] == 4)
    return S_MIFARE_AUTH;

  return S_CARD_ERROR;
}

/*
 * MifareUlWritePage
 * -----------------
 * Write a single 4-byte Mifare UltraLight page
 */
int MifareUlWritePage(unsigned char addr, unsigned char data[4])
{
  unsigned char cmd[8];
  unsigned char rsp[1];
  int r;

  if (data == NULL) return E_INTERNAL;

  /* Build the command */
  cmd[0] = 0x8F;    /* SPROX_PICCCOMMONWRITE */
  cmd[1] = 0xA2;    /* Mifare UL write command */
  cmd[2] = addr;    /* Address of page */
  cmd[3] = 4;       /* Size of data */

  /* Append data */
  memcpy(&cmd[4], data, 4);

  /* Send the command */
  r = Send(cmd, 8);
  if (r < 0) return r;

  /* Get the answer */
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] == 0)
    return 0; /* OK */

  /* Error ? */
  if (rsp[0] >= 26)
    return E_READER_FATAL;
  if (rsp[0] == 1)
    return S_NO_CARD;

  return S_CARD_ERROR;
}
