/*
 * SpringProx or K531 SDK :: direct/ref_direct.c
 * ---------------------------------------------
 *
 * Main function of the reference program for direct reader operation
 * (without SpringProx API)
 */

/*

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
  
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ref_direct.h"

int TraceReader = 0;
int TraceCard   = 0;

void Usage(void);
int GetFirmware(void);
const char *ErrorMessage(int err_code);

static char *SerialDevice = NULL;
static int   SerialAscii  = 0;  /* 0 -> binary, 1 -> ASCII */

int main(int argc, char **argv)
{
  int i;
  int RetCode = 0;
  int CmdCode = 0; /* 0 -> no action (show help and exit) */
                   /* 1 -> polling                        */
                   /* 2 -> mifare                         */
                   /* 3 -> mif_ul                         */
                   /* 4 -> tcl                            */
                   /* 5 -> st sr176                       */
                   
  for (i=1; i<argc; i++)
  {
    if ((i == (argc-1)) && (argv[i][0] != '-'))
    {
      /* Last argument */
      SerialDevice = argv[i];
      printf("Selecting device '%s'\n", SerialDevice);
    } else
    {
      /* Command line option */
      if (!strcmp(argv[i], "-poll"))
      {
        CmdCode = 1;
      } else
      if (!strcmp(argv[i], "-mif"))
      {
        CmdCode = 2;
      } else
      if (!strcmp(argv[i], "-mif_ul"))
      {
        CmdCode = 3;
      } else
      if (!strcmp(argv[i], "-tcl"))
      {
        CmdCode = 4;
      } else
      if (!strcmp(argv[i], "-sr176"))
      {
        CmdCode = 5;
      } else
      if (!strcmp(argv[i], "-cts"))
      {
        CmdCode = 6;
      } else
      if (!strcmp(argv[i], "-bin") || !strcmp(argv[i], "-binary"))
      {
        SerialAscii = 0;
      } else
      if (!strcmp(argv[i], "-asc") || !strcmp(argv[i], "-ascii"))
      {
        SerialAscii = 1;
      } else      
      if (!strcmp(argv[i], "-tr_rdr"))
      {
        TraceReader = 1;
      } else      
      if (!strcmp(argv[i], "-tr_crd"))
      {
        TraceCard = 1;
      } else      
      {
        printf("Unrecognized option '%s'\n", argv[i]);
        return EXIT_FAILURE;
      }   
    }
  }

  /* An action has been selected ? */
  if (CmdCode == 0)
  {
    Usage();
    return EXIT_FAILURE;
  }

  /* A communication device has been selected ? */
  if (SerialDevice == NULL)
  {
#ifdef WIN32
    printf("Selection COM1 as default\n");
    SerialDevice = "COM1";
#endif
#ifdef __linux
    printf("Selection /dev/ttyS0 as default\n");
    SerialDevice = "/dev/ttyS0";
#endif
    if (SerialDevice == NULL)
    {
      printf("Serial device not selected\n");
      return EXIT_FAILURE;
    }
  }

  /* Which protocol ? */
  if (SerialAscii)
    printf("Using ASCII protocol\n");
  else
    printf("Using 'fast' binary protocol\n");

  /*
   * OK, let's try to connect to the reader !
   * ----------------------------------------
   */

  printf("Connecting to reader...\n");

  /* Open the comm port */
  RetCode = SerialOpen(SerialDevice);
  if (RetCode != 0)
  {
    printf("Failed to open serial communication port\n");
    return EXIT_FAILURE;;
  }

  /* Get reader's firmware version to make sure it is available */
  RetCode = GetFirmware();
  if (RetCode == 0)
  {
    /*
     * OK, we may now run the specified function
     * -----------------------------------------
     */    
    switch (CmdCode)
    {   
      case 1  : RetCode = PollingProc(); break;
      case 2  : RetCode = MifareProc();  break;
      case 3  : RetCode = MifareUlProc();  break;
      case 4  : RetCode = TclProc();  break;
      case 5  : RetCode = StSr176Proc();  break;
      case 6  : RetCode = AskCtsProc();  break;
    
      default : printf("Internal error");
    }
  } else
  {
    /* GetFirmware failed */
    printf("Failed to establish communication with reader\n");
  }

  /* Close communication device */
  SerialClose();

  if (RetCode)
  {
    printf("Exiting with error %d\n\t(%s)\n", RetCode, ErrorMessage(RetCode));
    return EXIT_FAILURE;
  } else
  {
    printf("Exiting...\n");
    return EXIT_SUCCESS;
  }
}

const char *ErrorMessage(int err_code)
{
  switch (err_code)
  {
    case E_COMM_ERROR      : return "Communication error reported by OS";
    case E_READER_MUTE     : return "Reader is mute";
    case E_READER_STOP     : return "Reader's answer has been interrupted";
    case E_FRAME_BAD_CRC   : return "CRC error in reader's answer";
    case E_FRAME_BAD_LEN   : return "Reader's answer has not the expected length";
    case E_READER_FATAL    : return "Reader reported an internal or hardware error";
    case E_OVERFLOW        : return "Buffer overflow : reader's answer is too long for us";
    case E_INTERNAL        : return "Software internal error";
    case S_NO_CARD         : return "No card found, or card removed";
    case S_MIFARE_AUTH     : return "Access denied, Mifare authentication failed";
    case S_CARD_ERROR      : return "Reader <-> card communication error";
    case S_BAD_RESP_CODE   : return "Reader reported an error (not reader fault)";
    case S_BAD_RESP_SIZE   : return "Card's answer has not the expected length (not reader fault)";
  }
  return "";
}

void Usage(void)
{
  printf("Usage:\n");
  printf("\tref_direct -poll   [mode] [device] : ISO 14443 A & B polling\n");
  printf("\tref_direct -mif    [mode] [device] : Mifare Classic demo\n");
  printf("\tref_direct -mif_ul [mode] [device] : Mifare UltraLight demo\n");
  printf("\tref_direct -tcl    [mode] [device] : T=CL demo\n");
  printf("\tref_direct -sr176  [mode] [device] : ST SR176 demo\n");
  printf("\tref_direct -cts    [mode] [device] : ASK CTS256B or CTS512B demo\n");
  printf("\n");
  printf("[mode]   is either -ascii or -binary (default)\n");
  printf("[device] specified the communication port to be used\n");
  printf("\tWindows : COM1, COM2... default is COM1\n");
  printf("\tLinux   : /dev/ttyS0, /dev/ttyS1... default is /dev/ttyS0\n");
  printf("\n");
}


/*
 * Send
 * ----
 * The host-to-reader function
 */
int Send(unsigned char *Buffer, int Size)
{
  if (SerialAscii)
    return AsciiSend(Buffer, Size);
  return BinarySend(Buffer, Size);
}

/*
 * Recv
 * ----
 * The reader-to-host function
 */
int Recv(unsigned char *Buffer, int Size)
{
  if (SerialAscii)
    return AsciiRecv(Buffer, Size);
  return BinaryRecv(Buffer, Size);
}



int GetFirmware(void)
{
  unsigned char cmd[1];
  unsigned char rsp[32 + 1];
  int r;

  /* Build the command */
  cmd[0] = 0x4F;

  /* Send the command */
  r = Send(cmd, 1);
  if (r < 0) return r;

  /* Check we receive something */
  r = Recv(rsp, sizeof(rsp));
  if (r < 0) return r;

  if (rsp[0] != 0) return S_BAD_RESP_CODE;
  if (r < 4+1)     return E_FRAME_BAD_LEN;

  /* No error, this is the firmware */
  printf("Connected to %c%c%c%c\n", rsp[1], rsp[2], rsp[3], rsp[4]);
  return 0;
}
